@@ -1,18 +1,26 @@
 import copy
 import inspect
-from typing import Union, Optional, Callable, Any, Awaitable
+from typing import Union, Optional, Callable, Any, NoReturn
+from collections.abc import Awaitable
 
 import httpx
 
 from pytest_httpx import _httpx_internals
+from pytest_httpx._options import _HTTPXMockOptions
 from pytest_httpx._pretty_print import RequestDescription
 from pytest_httpx._request_matcher import _RequestMatcher
 
 
 class HTTPXMock:
-    def __init__(self) -> None:
+    """
+    This class is only exposed for `httpx_mock` fixture type hinting purpose.
+    """
+
+    def __init__(self, options: _HTTPXMockOptions) -> None:
+        """Private and subject to breaking changes without notice."""
+        self._options = options
         self._requests: list[
-            tuple[Union[httpx.BaseTransport, httpx.AsyncBaseTransport], httpx.Request]
+            tuple[Union[httpx.HTTPTransport, httpx.AsyncHTTPTransport], httpx.Request]
         ] = []
         self._callbacks: list[
             tuple[
@@ -25,6 +33,7 @@ class HTTPXMock:
                 ],
             ]
         ] = []
+        self._requests_not_matched: list[httpx.Request] = []
 
     def add_response(
         self,
@@ -57,6 +66,10 @@ class HTTPXMock:
         :param match_headers: HTTP headers identifying the request(s) to match. Must be a dictionary.
         :param match_content: Full HTTP body identifying the request(s) to match. Must be bytes.
         :param match_json: JSON decoded HTTP body identifying the request(s) to match. Must be JSON encodable.
+        :param match_data: Multipart data (excluding files) identifying the request(s) to match. Must be a dictionary.
+        :param match_files: Multipart files identifying the request(s) to match. Refer to httpx documentation for more information on supported values: https://www.python-httpx.org/advanced/clients/#multipart-file-encoding
+        :param is_optional: True will mark this response as optional, False will expect a request matching it. Must be a boolean. Default to the opposite of assert_all_responses_were_requested option value (itself defaulting to True, meaning this parameter default to False).
+        :param is_reusable: True will allow re-using this response even if it already matched, False prevent re-using it. Must be a boolean. Default to the can_send_already_matched_responses option value (itself defaulting to False).
         """
 
         json = copy.deepcopy(json) if json is not None else None
@@ -96,8 +109,13 @@ class HTTPXMock:
         :param match_headers: HTTP headers identifying the request(s) to match. Must be a dictionary.
         :param match_content: Full HTTP body identifying the request(s) to match. Must be bytes.
         :param match_json: JSON decoded HTTP body identifying the request(s) to match. Must be JSON encodable.
+        :param match_data: Multipart data (excluding files) identifying the request(s) to match. Must be a dictionary.
+        :param match_files: Multipart files identifying the request(s) to match. Refer to httpx documentation for more information on supported values: https://www.python-httpx.org/advanced/clients/#multipart-file-encoding
+        :param match_extensions: Extensions identifying the request(s) to match. Must be a dictionary.
+        :param is_optional: True will mark this callback as optional, False will expect a request matching it. Must be a boolean. Default to the opposite of assert_all_responses_were_requested option value (itself defaulting to True, meaning this parameter default to False).
+        :param is_reusable: True will allow re-using this callback even if it already matched, False prevent re-using it. Must be a boolean. Default to the can_send_already_matched_responses option value (itself defaulting to False).
         """
-        self._callbacks.append((_RequestMatcher(**matchers), callback))
+        self._callbacks.append((_RequestMatcher(self._options, **matchers), callback))
 
     def add_exception(self, exception: Exception, **matchers: Any) -> None:
         """
@@ -112,6 +130,11 @@ class HTTPXMock:
         :param match_headers: HTTP headers identifying the request(s) to match. Must be a dictionary.
         :param match_content: Full HTTP body identifying the request(s) to match. Must be bytes.
         :param match_json: JSON decoded HTTP body identifying the request(s) to match. Must be JSON encodable.
+        :param match_data: Multipart data (excluding files) identifying the request(s) to match. Must be a dictionary.
+        :param match_files: Multipart files identifying the request(s) to match. Refer to httpx documentation for more information on supported values: https://www.python-httpx.org/advanced/clients/#multipart-file-encoding
+        :param match_extensions: Extensions identifying the request(s) to match. Must be a dictionary.
+        :param is_optional: True will mark this exception response as optional, False will expect a request matching it. Must be a boolean. Default to the opposite of assert_all_responses_were_requested option value (itself defaulting to True, meaning this parameter default to False).
+        :param is_reusable: True will allow re-using this exception response even if it already matched, False prevent re-using it. Must be a boolean. Default to the can_send_already_matched_responses option value (itself defaulting to False).
         """
 
         def exception_callback(request: httpx.Request) -> None:
@@ -123,9 +146,11 @@ class HTTPXMock:
 
     def _handle_request(
         self,
-        real_transport: httpx.BaseTransport,
+        real_transport: httpx.HTTPTransport,
         request: httpx.Request,
     ) -> httpx.Response:
+        # Store the content in request for future matching
+        request.read()
         self._requests.append((real_transport, request))
 
         callback = self._get_callback(real_transport, request)
@@ -135,16 +160,15 @@ class HTTPXMock:
             if response:
                 return _unread(response)
 
-        raise httpx.TimeoutException(
-            self._explain_that_no_response_was_found(real_transport, request),
-            request=request,
-        )
+        self._request_not_matched(real_transport, request)
 
     async def _handle_async_request(
         self,
-        real_transport: httpx.AsyncBaseTransport,
+        real_transport: httpx.AsyncHTTPTransport,
         request: httpx.Request,
     ) -> httpx.Response:
+        # Store the content in request for future matching
+        await request.aread()
         self._requests.append((real_transport, request))
 
         callback = self._get_callback(real_transport, request)
@@ -156,6 +180,14 @@ class HTTPXMock:
                     response = await response
                 return _unread(response)
 
+        self._request_not_matched(real_transport, request)
+
+    def _request_not_matched(
+        self,
+        real_transport: Union[httpx.AsyncHTTPTransport, httpx.HTTPTransport],
+        request: httpx.Request,
+    ) -> NoReturn:
+        self._requests_not_matched.append(request)
         raise httpx.TimeoutException(
             self._explain_that_no_response_was_found(real_transport, request),
             request=request,
@@ -170,15 +202,29 @@ class HTTPXMock:
 
         message = f"No response can be found for {RequestDescription(real_transport, request, matchers)}"
 
-        matchers_description = "\n".join([str(matcher) for matcher in matchers])
+        already_matched = []
+        unmatched = []
+        for matcher in matchers:
+            if matcher.nb_calls:
+                already_matched.append(matcher)
+            else:
+                unmatched.append(matcher)
+
+        matchers_description = "\n".join(
+            [f"- {matcher}" for matcher in unmatched + already_matched]
+        )
         if matchers_description:
             message += f" amongst:\n{matchers_description}"
+            # If we could not find a response, but we have already matched responses
+            # it might be that user is expecting one of those responses to be reused
+            if any(not matcher.is_reusable for matcher in already_matched):
+                message += "\n\nIf you wanted to reuse an already matched response instead of registering it again, refer to https://github.com/Colin-b/pytest_httpx/blob/master/README.md#allow-to-register-a-response-for-more-than-one-request"
 
         return message
 
     def _get_callback(
         self,
-        real_transport: Union[httpx.BaseTransport, httpx.AsyncBaseTransport],
+        real_transport: Union[httpx.HTTPTransport, httpx.AsyncHTTPTransport],
         request: httpx.Request,
     ) -> Optional[
         Callable[
@@ -203,9 +249,13 @@ class HTTPXMock:
                 matcher.nb_calls += 1
                 return callback
 
-        # Or the last registered
-        matcher.nb_calls += 1
-        return callback
+        # Or the last registered (if it can be reused)
+        if matcher.is_reusable:
+            matcher.nb_calls += 1
+            return callback
+
+        # All callbacks have already been matched and last registered cannot be reused
+        return None
 
     def get_requests(self, **matchers: Any) -> list[httpx.Request]:
         """
@@ -219,8 +269,11 @@ class HTTPXMock:
         :param match_headers: HTTP headers identifying the requests to retrieve. Must be a dictionary.
         :param match_content: Full HTTP body identifying the requests to retrieve. Must be bytes.
         :param match_json: JSON decoded HTTP body identifying the requests to retrieve. Must be JSON encodable.
+        :param match_data: Multipart data (excluding files) identifying the requests to retrieve. Must be a dictionary.
+        :param match_files: Multipart files identifying the requests to retrieve. Refer to httpx documentation for more information on supported values: https://www.python-httpx.org/advanced/clients/#multipart-file-encoding
+        :param match_extensions: Extensions identifying the requests to retrieve. Must be a dictionary.
         """
-        matcher = _RequestMatcher(**matchers)
+        matcher = _RequestMatcher(self._options, **matchers)
         return [
             request
             for real_transport, request in self._requests
@@ -239,49 +292,50 @@ class HTTPXMock:
         :param match_headers: HTTP headers identifying the request to retrieve. Must be a dictionary.
         :param match_content: Full HTTP body identifying the request to retrieve. Must be bytes.
         :param match_json: JSON decoded HTTP body identifying the request to retrieve. Must be JSON encodable.
+        :param match_data: Multipart data (excluding files) identifying the request to retrieve. Must be a dictionary.
+        :param match_files: Multipart files identifying the request to retrieve. Refer to httpx documentation for more information on supported values: https://www.python-httpx.org/advanced/clients/#multipart-file-encoding
+        :param match_extensions: Extensions identifying the request to retrieve. Must be a dictionary.
         :raises AssertionError: in case more than one request match.
         """
         requests = self.get_requests(**matchers)
         assert (
             len(requests) <= 1
-        ), f"More than one request ({len(requests)}) matched, use get_requests instead."
+        ), f"More than one request ({len(requests)}) matched, use get_requests instead or refine your filters."
         return requests[0] if requests else None
 
-    def reset(self, assert_all_responses_were_requested: bool) -> None:
+    def reset(self) -> None:
         self._requests.clear()
-        not_called = self._reset_callbacks()
-
-        if assert_all_responses_were_requested:
-            matchers_description = "\n".join([str(matcher) for matcher in not_called])
-
-            assert (
-                not not_called
-            ), f"The following responses are mocked but not requested:\n{matchers_description}"
-
-    def _reset_callbacks(self) -> list[_RequestMatcher]:
-        callbacks_not_executed = [
-            matcher for matcher, _ in self._callbacks if not matcher.nb_calls
-        ]
         self._callbacks.clear()
-        return callbacks_not_executed
+        self._requests_not_matched.clear()
 
+    def _assert_options(self) -> None:
+        callbacks_not_executed = [
+            matcher for matcher, _ in self._callbacks if matcher.should_have_matched()
+        ]
+        matchers_description = "\n".join(
+            [f"- {matcher}" for matcher in callbacks_not_executed]
+        )
 
-class _PytestSyncTransport(httpx.BaseTransport):
-    def __init__(self, real_transport: httpx.BaseTransport, mock: HTTPXMock):
-        self._real_transport = real_transport
-        self._mock = mock
+        assert not callbacks_not_executed, (
+            "The following responses are mocked but not requested:\n"
+            f"{matchers_description}\n"
+            "\n"
+            "If this is on purpose, refer to https://github.com/Colin-b/pytest_httpx/blob/master/README.md#allow-to-register-more-responses-than-what-will-be-requested"
+        )
 
-    def handle_request(self, request: httpx.Request) -> httpx.Response:
-        return self._mock._handle_request(self._real_transport, request)
-
-
-class _PytestAsyncTransport(httpx.AsyncBaseTransport):
-    def __init__(self, real_transport: httpx.AsyncBaseTransport, mock: HTTPXMock):
-        self._real_transport = real_transport
-        self._mock = mock
-
-    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
-        return await self._mock._handle_async_request(self._real_transport, request)
+        if self._options.assert_all_requests_were_expected:
+            requests_description = "\n".join(
+                [
+                    f"- {request.method} request on {request.url}"
+                    for request in self._requests_not_matched
+                ]
+            )
+            assert not self._requests_not_matched, (
+                f"The following requests were not expected:\n"
+                f"{requests_description}\n"
+                "\n"
+                "If this is on purpose, refer to https://github.com/Colin-b/pytest_httpx/blob/master/README.md#allow-to-not-register-responses-for-every-request"
+            )
 
 
 def _unread(response: httpx.Response) -> httpx.Response:
